/*
 *  face_detect.cpp
 *  $Id$
 *
 *  Author
 *  Roman Juranek <ijuranek@fit.vutbr.cz>
 *
 *  Graph@FIT
 *  Department of Computer Graphics and Multimedia
 *  Faculty of Information Technology
 *  Brno University of Technology
 *
 *  Description
 *  Example of multi-scale detection of faces
 *
 */


#include <libabr.h> // libabr interface
#include "frontal-face-lbp2x2-a20.h" // face detection classifier

// OpenCV (camera and image representation)
#include <cv.h>
#include <cxcore.h>
#include <highgui.h>

// STL stuff
#include <iostream>
#include <numeric>

// OpenMP
#include <omp.h>

using namespace std;


int main(int argc, char ** argv)
{
    // Open camera
    CvCapture * cap = cvCaptureFromCAM(0);
    //CvCapture * cap = cvCreateFileCapture("v4l:///dev/video0");

    // Exit on 'no camera'
    if (!cap) return 0;

    // Alloc images of propper sizes
    IplImage * src = cvQueryFrame(cap);
    CvSize base_sz = cvGetSize(src);
    IplImage * gray = cvCreateImage(base_sz, IPL_DEPTH_8U, 1);

    // Downscale of input (2 means that detection is performed on 2x downscaled input)
    float scale=1;
    
    // Initialize preprocessing module
    init_preprocess();

    // Create structures for image preprocesing
    ImageLayout * il = create_layout(cvSize(base_sz.width/scale,base_sz.height/scale), 1.25, cvSize(28,28));
    PreprocessedImage * pp = create_preprocessed_image(cvGetSize(il->image));

    // Set classifier. It could by also loaded by load_classifier_XML()
    TClassifier * c = &frontal_face_lbp2x2_a20;
    
    // Initialize the structure
    init_classifier(c);

    // Set up parameters
    c->threshold = 25;

    // Alloc results
    Detection results[1000];

    // Set up scan parameters
    ScanParams sp;
    sp.use_lap_prunning = 0; // 1 for laplace prunning which may cause some speed-up
    sp.lap_threshold = 0.005;
    sp.start_stage = 0;
    sp.end_stage = c->stage_count;
    sp.response_map = cvCreateImage(il->total_sz, IPL_DEPTH_32F, 1);

    cvNamedWindow("IMG");
    
    bool fin = false;

    omp_set_num_threads(4);

    while (((src = cvQueryFrame(cap)) != 0) && !fin)
    {
        cvCvtColor(src, gray, CV_RGB2GRAY);
        
         //int n = detect_objects(gray, il, pp, c, &sp, scan_image_lbp, results, results+1000, PP_LBP_IMAGE, 0);
         

        cvZero(sp.response_map);
        insert_image_layout(il, gray);
        preprocess_image(il->image, pp, PP_LBP_IMAGE);
        sp.start_stage = 0;
        sp.end_stage = 10;
        prepare_classifier(c, pp, RECALC_OFFSET);
        scan_image_intensity_map(pp, c, &sp, 0, 0, 0);
        sp.start_stage = 10;
        sp.end_stage = c->stage_count;
        prepare_classifier(c, pp, RECALC_RANKS);
        int n0 = scan_image_lbp(pp, c, &sp, results, results+1000, 0);
        int n = unlayout_results(results, results+n0, il, 1, 1);
       
        
        // Draw results to the image
        for (Detection * r = results; r != results + n; ++r)
        {
            cvRectangle(src, cvPoint(r->x, r->y), cvPoint(r->x+r->width, r->y+r->height), CV_RGB(0,0,0), 3);
            cvRectangle(src, cvPoint(r->x, r->y), cvPoint(r->x+r->width, r->y+r->height), CV_RGB(255,255,255));
        }

        cvShowImage("IMG", src);
        cvShowImage("MAP", sp.response_map);

        // Keyboard
        int c = cvWaitKey(5) & 0xFF;
        switch (c)
        {
        case 'q':
        case 'Q':
            fin = true;
            break;
        }
    }

    cvReleaseCapture(&cap);
    cvReleaseImage(&gray);
    destroy_layout(&il);
    release_preprocessed_image(&pp);

    return 0;
}

