#include "core.h"
#include "const.h"
#include <vector>
#include <cstdio>

using namespace std;

// Shuffle alphas for LRP
int init_classifier(TClassifier * c)
{
    float * stg_alpha = c->alpha;

    for (unsigned s = 0; s < c->stage_count; ++s, stg_alpha += c->alpha_count)
    {
        TStage * stage = c->stage + s;

        if (stage->w > 2 || stage->h > 2)
            return 0;

        // get feature type
        // 000000hw
        stage->alpha = stg_alpha;
        stage->sz_type = ((stage->h-1) << 1) | (stage->w-1);
        stage->pos_type = ((stage->y & 0x03) << 2) | (stage->x & 0x03);
        stage->offset = 0;
    }

    return 1;
}


void prepare_classifier(TClassifier * c, PreprocessedImage * PI, int options)
{
    if (options == NONE)
        return;

    if ((options & RECALC_OFFSET))
    {
        IplImage * img;
        int px_sz;
        if (options & OFFSET_INTEGRAL)
        {
            img = &(PI->integral);
            px_sz = sizeof(int);
        }
        else
        {
            img = &(PI->intensity);
            px_sz = sizeof(char);
        }
        for (unsigned s = 0; s < c->stage_count; ++s)
        {
            TStage * stage = (TStage*)c->stage + s;
            stage->offset = stage->y * img->widthStep + (stage->x * px_sz);
        }
    }

    if ((options & RECALC_RANKS) && (c->tp == LRD || c->tp == LRP))
    {
        assert(c->ranks != 0);

        for (unsigned s = 0; s < c->stage_count; ++s)
        {
            const TStage & stage = c->stage[s];
            int * ranks = c->ranks + 8 * s;

            for (int t = 0; t < 4; ++t)
            {
                int & A = ranks[2 * t + 0];
                int & B = ranks[2 * t + 1];
                A = rank_table[t][int(stage.A)];
                B = rank_table[t][int(stage.B)];
                if (A >= 8) A += PI->irow_size[int(stage.sz_type)] - 8;
                if (B >= 8) B += PI->irow_size[int(stage.sz_type)] - 8;
            }
        } // stages
    } // if
}

int detect_objects(
        IplImage * src,
        ImageLayout * IL,
        PreprocessedImage * PI,
        TClassifier * c,
        ScanParams * sp,
        ScanImageFunc scan_image,
        Detection * first, Detection * last,
        int pi_opts,
        int * hist)
{
    // Asserts?
    insert_image_layout(IL, src);
    preprocess_image(IL->image, PI, pi_opts);
    int raw_dets = scan_image(PI, c, sp, first, last, hist);
    int valid_dets = 0;
    if (raw_dets > 0)
    {
        //cerr << src->width << "," << IL->sz.width;
        float scale_x = src->width / IL->sz.width; 
        float scale_y = src->height / IL->sz.height;
        valid_dets = unlayout_results(first, first+raw_dets, IL, scale_x, scale_y);
        // if (group_threshold > 0)
        //  group_results();
    }
    return valid_dets;
}


void prepare_laplace_prunning(IplImage * lap_integral, ScanParams * sp, TClassifier * c, char ** base, int * x_ofs, int * y_ofs, int * threshold)
{
    int w_size = c->width * c->height;
    *x_ofs = c->width;
    *y_ofs = c->height * lap_integral->widthStep; // bytes
    *threshold = sp->lap_threshold * 1024 * w_size; 
    *base = lap_integral->imageData;
}

int test_laplace(char * ci_base, int x_ofs, int y_ofs)
{
    // test if the window should be evaluated
    char * base = ci_base;
    int sum = 0;
    sum += *(int*)(base);
    sum -= *((int*)(base) + x_ofs);
    base += y_ofs;
    sum -= *(int*)(base);
    sum += *((int*)(base) + x_ofs);
    return sum;
}


int unlayout_results(Detection * first, Detection * last, ImageLayout * IL, float scale_x, float scale_y)
{
    int max_candidates = IL->sub_area.size();
    
    bool processed[last - first];
    fill(processed, processed+(last-first), false);
    
    Detection * d = first;
    while (d != last)
    {
        for (int i = 0; i < max_candidates; ++i)
        {
            CvRect & r = IL->sub_area[i];
            if (d->x > r.x && d->y > r.y &&
               (d->x + d->width < r.x + r.width) && (d->y + d->height < r.y + r.height))
            {
                float sx = IL->scale[i] * scale_x;
                float sy = IL->scale[i] * scale_y;
                // Subtract subarea position
                d->x -= r.x;
                d->y -= r.y;
                // Apply subarea scale
                d->x *= sx;
                d->y *= sy;
                d->width *= sx;
                d->height *= sy;

                processed[d - first] = true;
            }

        }
        ++d;
    }

    // TODO remove unprocessed results!

    return last - first;
}

int group_results(Detection * first, Detection * last, float overlap)
{
    return 0;
}

